@@ -6,16 +6,24 @@
 
 __author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"
 
-import argparse, logging, os, pickle, sys, time, ctypes
+import argparse
+import logging
+import os
+import pickle
+import sys
+import time
+import ctypes
 import subprocess
 
 import model
-
-from memory_mapper import MemoryMapper
+from haystack.memory_mapper import MemoryMapper as MemoryMapper
 
 log=logging.getLogger('haystack')
 
-environSep = ':'
+if not sys.platform.startswith('win'):
+  environSep = ':'
+else:
+  environSep = ';'
 
 
 
@@ -28,9 +36,8 @@ class StructFinder:
   
   @param mappings: address space
   @param targetMappings: search perimeter. If None, all mappings are used in the search perimeter.
-  
   '''
-  def __init__(self, mappings, targetMappings=None):
+  def __init__(self, mappings, targetMappings=None, updateCb=None):
     self.mappings = mappings
     if type(mappings) == bool:
       raise TypeError()
@@ -39,8 +46,9 @@ class StructFinder:
       self.targetMappings = mappings
     log.debug('StructFinder on %d memorymappings. Search Perimeter on %d mappings.'%(len(self.mappings), len(self.targetMappings)) )
     return
-
-  def find_struct(self, struct, hintOffset=0, maxNum = 10, maxDepth=10 ):
+    
+  def find_struct(self, structType, hintOffset=0, maxNum = 10, maxDepth=10 ):
+    """ Iterate on all targetMappings to find a structure. """
     log.warning("Restricting search to %d memory mapping."%(len(self.targetMappings)))
     outputs=[]
     for m in self.targetMappings:
@@ -51,8 +59,8 @@ class StructFinder:
         continue
       else:
         log.debug("%s,%s"%(m,m.permissions))
-      log.debug('look for %s'%(struct))
-      outputs.extend(self.find_struct_in( m, struct, hintOffset=hintOffset, maxNum=maxNum, maxDepth=maxDepth))
+      log.debug('look for %s'%(structType))
+      outputs.extend(self.find_struct_in( m, structType, hintOffset=hintOffset, maxNum=maxNum, maxDepth=maxDepth))
       # check out
       if len(outputs) >= maxNum:
         log.debug('Found enough instance. returning results.')
@@ -60,14 +68,14 @@ class StructFinder:
     # if we mmap, we could yield
     return outputs
 
-  def find_struct_in(self, memoryMap, struct, hintOffset=0, maxNum=10, maxDepth=99 ):
+  def find_struct_in(self, memoryMap, structType, hintOffset=0, maxNum=10, maxDepth=99 ):
     '''
-      Looks for struct in memory, using :
-        hints from struct (default values, and such)
-        guessing validation with instance(struct)().isValid()
-        and confirming with instance(struct)().loadMembers()
+      Looks for structType instances in memory, using :
+        hints from structType (default values, and such)
+        guessing validation with instance(structType)().isValid()
+        and confirming with instance(structType)().loadMembers()
       
-      returns POINTERS to struct.
+      returns POINTERS to structType instances.
     '''
 
     # update process mappings
@@ -77,7 +85,7 @@ class StructFinder:
     start=memoryMap.start  
     end=memoryMap.end
     plen=ctypes.sizeof(ctypes.c_void_p) # use aligned words only
-    structlen=ctypes.sizeof(struct)
+    structlen=ctypes.sizeof(structType)
     #ret vals
     outputs=[]
     # alignement
@@ -88,7 +96,7 @@ class StructFinder:
       align=hintOffset%plen
       start=start+ (hintOffset-align)
      
-    # parse for struct on each aligned word
+    # parse for structType on each aligned word
     log.debug("checking 0x%lx-0x%lx by increment of %d"%(start, (end-structlen), plen))
     instance=None
     import time
@@ -101,7 +109,7 @@ class StructFinder:
         log.debug('processed %d bytes  - %02.02f test/sec'%(p2, (p2-p)/(plen*(time.time()-t0)) ))
         t0=time.time()
         p=p2
-      instance,validated= self.loadAt( memoryMap, offset, struct, maxDepth) 
+      instance,validated= self.loadAt( memoryMap, offset, structType, maxDepth) 
       if validated:
         log.debug( "found instance @ 0x%lx"%(offset) )
         # do stuff with it.
@@ -112,12 +120,16 @@ class StructFinder:
     return outputs
 
 
-  def loadAt(self, memoryMap, offset, struct, depth=99 ):
-    log.debug("Loading %s from 0x%lx "%(struct,offset))
-    instance=struct.from_buffer_copy(memoryMap.readStruct(offset,struct))
+  def loadAt(self, memoryMap, offset, structType, depth=99 ):
+    ''' 
+      loads a haystack ctypes structure from a specific offset. 
+        return (instance,validated) with instance being the haystack ctypes structure instance and validated a boolean True/False.
+    '''
+    log.debug("Loading %s from 0x%lx "%(structType,offset))
+    instance=structType.from_buffer_copy(memoryMap.readStruct(offset,structType))
     # check if data matches
     if ( instance.loadMembers(self.mappings, depth) ):
-      log.info( "found instance %s @ 0x%lx"%(struct,offset) )
+      log.info( "found instance %s @ 0x%lx"%(structType,offset) )
       # do stuff with it.
       validated=True
     else:
@@ -125,6 +137,37 @@ class StructFinder:
       validated=False
     return instance,validated
 
+
+class VerboseStructFinder(StructFinder):
+  ''' structure finder with a update callback to be more verbose.
+  Will search a structure defined by it's pointer and other constraints.
+  Address space is defined by  mappings.
+  Target memory perimeter is defined by targetMappings.
+  targetMappings is included in mappings.
+  
+  @param mappings: address space
+  @param targetMappings: search perimeter. If None, all mappings are used in the search perimeter.
+  @param updateCb: callback func. for periodic status update
+  '''
+  def __init__(self,mappings, targetMappings=None, updateCb=None):
+    StructFinder.__init__(self,mappings, targetMappings)
+    self.updateCb = updateCb
+    self._updateCb_init()
+    
+  def _updateCb_init(self):
+    # approximation
+    nb = lambda x : ((x.end-x.start)/4)
+    self._update_nb_steps = sum([nb(m) for m in self.targetMappings])
+    self._update_i = 0
+
+  def loadAt(self, memoryMap, offset, structType, depth=99 ):
+    self._update_i+=1
+    self.updateCb(self._update_i)
+    StructFinder.loadAt(memoryMap, offset, structType, depth=depth )
+
+
+
+
 def hasValidPermissions(memmap):
   ''' memmap must be 'rw..' or shared '...s' '''
   perms=memmap.permissions
@@ -132,6 +175,7 @@ def hasValidPermissions(memmap):
 
 
 def _callFinder(cmd_line):
+  """ Call the haystack finder in a subprocess. Will use pickled objects to communicate results. """
   env = os.environ
   env['PYTHONPATH'] = environSep.join(sys.path) # add possible pythonpaths to environnement
   p = subprocess.Popen(cmd_line, stdin=None, stdout=subprocess.PIPE, close_fds=True , env=env)
@@ -145,6 +189,11 @@ def getMainFile():
 
 
 def checkModulePath(typ):
+  '''
+    add typ module's path to sys.path
+    If the type is a generated haystack structure type, 
+    dump the '_generated' string from the module name and import it under the new module name.
+  '''
   name = typ.__name__
   module,sep,kname = name.rpartition('.')
   # add the __file__ module to sys.path for it to be reachable by subprocess
@@ -158,23 +207,38 @@ def checkModulePath(typ):
     try:
       plainmod = typ.__module__.replace('_generated','')
       mod = __import__( plainmod, globals(), locals(), [kname])
-      struct = '.'.join([plainmod , kname])
-      log.info('trying %s instead of %s'%(struct, name))
-      return struct
+      structName = '.'.join([plainmod , kname])
+      log.info('trying %s instead of %s'%(structName, name))
+      return structName
     except ImportError:
       # shhh  
       pass
-  struct = '.'.join([typ.__module__,typ.__name__]) # we pass a str anyway...
-  return struct
+  structName = '.'.join([typ.__module__,typ.__name__]) # we pass a str anyway...
+  return structName
 
-def _findStruct(pid=None, memfile=None, memdump=None, struct=None, maxNum=1, 
-              fullScan=False, nommap=False, hint=None, debug=None ):
-  ''' '''
-  if type(struct) != type(ctypes.Structure):
-    raise TypeError('struct arg must be a ctypes.Structure')
-  structname = checkModulePath(struct) # add to sys.path
-  cmd_line=[sys.executable, getMainFile(), "%s"%structname]
-  if debug:
+def _findStruct(pid=None, memfile=None, memdump=None, structType=None, maxNum=1, 
+              fullScan=False, nommap=False, hint=None, debug=None, quiet=True ):
+  ''' 
+    Find all occurences of a specific structure from a process memory.
+    Returns occurences as objects.
+
+    Call a subprocess to ptrace a process. That way, self.process is not attached to the target PID by any means.
+    
+    @param pid is the process PID.
+    @param memfile the file containing a direct dump of the memory mapping ( optionnal)
+    @param memdump the file containing a memory dump 
+    @param structType the structure type.
+    @param offset the offset from which the structure must be loaded.
+    @param debug if True, activate debug logs.
+    @param maxNum the maximum number of expected results. Searching will stop after that many findings. -1 is unlimited.
+  '''
+  if type(structType) != type(ctypes.Structure):
+    raise TypeError('structType arg must be a ctypes.Structure')
+  structName = checkModulePath(structType) # add to sys.path
+  cmd_line=[sys.executable, getMainFile(), "%s"%structName]
+  if quiet:
+    cmd_line.insert(2,"--quiet")
+  elif debug:
     cmd_line.insert(2,"--debug")
   if nommap:
     cmd_line.insert(2,'--nommap')
@@ -194,24 +258,52 @@ def _findStruct(pid=None, memfile=None, memdump=None, struct=None, maxNum=1,
   # call me
   outs=_callFinder(cmd_line)
   if len(outs) == 0:
-    log.error("The %s has not been found."%(struct))
+    log.error("The %s has not been found."%(structName))
     return None
   #
   return outs
 
-def findStruct(pid, struct, maxNum=1, fullScan=False, nommap=False, debug=False):
-  return _findStruct(pid=pid, struct=struct, maxNum=maxNum, fullScan=fullScan, nommap=nommap, debug=debug)
+def findStruct(pid, structType, maxNum=1, fullScan=False, nommap=False, debug=False, quiet=True):
+  ''' 
+    Find all occurences of a specific structure from a process memory.
+    
+    @param pid is the process PID.
+    @param structType the structure type.
+    @param maxNum the maximum number of expected results. Searching will stop after that many findings. -1 is unlimited.
+    @param fullScan obselete
+    @param nommap if True, do not use mmap while searching.
+    @param debug if True, activate debug logs.
+  '''
+  return _findStruct(pid=pid, structType=structType, maxNum=maxNum, fullScan=fullScan, nommap=nommap, debug=debug, quiet=quiet)
   
-def findStructInFile(filename, struct, hint=None, maxNum=1, fullScan=False, debug=False):
-  return _findStruct(memfile=filename, struct=struct, maxNum=maxNum, fullScan=fullScan, debug=debug)
+def findStructInFile(filename, structType, hint=None, maxNum=1, fullScan=False, debug=False, quiet=True):
+  ''' 
+    Find all occurences of a specific structure from a process memory in a file.
+    
+    @param filename is the file containing the memory mapping content.
+    @param structType the structure type.
+    @param maxNum the maximum number of expected results. Searching will stop after that many findings. -1 is unlimited.
+    @param hint obselete
+    @param fullScan obselete
+    @param debug if True, activate debug logs.
+  '''
+  return _findStruct(memfile=filename, structType=structType, maxNum=maxNum, fullScan=fullScan, debug=debug, quiet=quiet)
 
 
-def refreshStruct(pid, struct, offset, debug=False, nommap=False):
-  ''' '''
-  if type(struct) != type(ctypes.Structure):
-    raise TypeError('struct arg must be a ctypes.Structure')
-  structname = checkModulePath(struct) # add to sys.path
-  cmd_line=[sys.executable, getMainFile(),  '%s'%structname]
+def refreshStruct(pid, structType, offset, debug=False, nommap=False):
+  ''' 
+    returns the pickled or text representation of a structure, from a given offset in a process memory.
+    
+    @param pid is the process PID.
+    @param structType the structure Type.
+    @param offset the offset from which the structure must be loaded.
+    @param debug if True, activate debug logs.
+    @param nommap if True, do not use mmap when mapping the memory
+  '''
+  if type(structType) != type(ctypes.Structure):
+    raise TypeError('structType arg must be a ctypes.Structure')
+  structName = checkModulePath(structType) # add to sys.path
+  cmd_line=[sys.executable, getMainFile(),  '%s'%structName]
   if debug:
     cmd_line.insert(2,"--debug")
   if nommap:
@@ -231,18 +323,19 @@ def refreshStruct(pid, struct, offset, debug=False, nommap=False):
     return None,None
   return instance,offset
 
-
-def usage(parser):
-  parser.print_help()
-  sys.exit(-1)
-  
 def argparser():
+  """
+    Builds the argparse tree.
+    See the command line --help .
+  """
   rootparser = argparse.ArgumentParser(prog='StructFinder', description='Parse memory structs and pickle them.')
   rootparser.add_argument('--string', dest='human', action='store_const', const=True, help='Print results as human readable string')
   rootparser.add_argument('--debug', dest='debug', action='store_const', const=True, help='setLevel to DEBUG')
+  rootparser.add_argument('--quiet', dest='quiet', action='store_const', const=True, help='setLevel to ERROR only')
   rootparser.add_argument('--interactive', dest='interactive', action='store_const', const=True, help='drop to python command line after action')
   rootparser.add_argument('--nommap', dest='mmap', action='store_const', const=False, default=True, help='disable mmap()-ing')
-  rootparser.add_argument('structType', type=str, help='Structure type name')
+  rootparser.add_argument('structName', type=str, help='Structure type name')
+  rootparser.add_argument('--baseOffset', type=str, help='base offset of the memory map in the dump file.')
   
   target = rootparser.add_mutually_exclusive_group(required=True)
   target.add_argument('--pid', type=int, help='Target PID')
@@ -264,6 +357,12 @@ def argparser():
 
 
 def getKlass(name):
+  '''
+    Returns the class type from a structure name.
+    The class' module is dynamically loaded.
+    
+    @param name a haystack structure's text name. ( 'sslsnoop.ctypes_openssh.session_state' for example )
+  '''
   module,sep,kname=name.rpartition('.')
   mod = __import__(module, globals(), locals(), [kname])
   klass = getattr(mod, kname)  
@@ -274,11 +373,22 @@ def getKlass(name):
   #log.error(getattr(mod, kname+'_py'))
   return klass
 
-def searchIn(structType, mappings, targetMappings=None, maxNum=-1):
-  log.debug('searchIn: %s - %s'%(structType,mappings))
-  structType = getKlass(structType)
+def searchIn(structName, mappings, targetMappings=None, maxNum=-1):
+  """
+    Search a structure in a specific memory mapping.
+    
+    if targetMappings is not specified, the search will occur in each memory mappings
+     in mappings.
+    
+    @param structName the structure name.
+    @param mappings the memory mappings list.
+    @param targetMappings the list of specific mapping to look into.
+    @param maxNum the maximum number of results expected. -1 for infinite.
+  """
+  log.debug('searchIn: %s - %s'%(structName,mappings))
+  structType = getKlass(structName)
   finder = StructFinder(mappings, targetMappings)
-  # find all possible struct instance
+  # find all possible structType instance
   outs=finder.find_struct( structType, maxNum=maxNum)
   # prepare outputs
   ret=[ (ss.toPyObject(),addr) for ss, addr in outs]
@@ -289,8 +399,17 @@ def searchIn(structType, mappings, targetMappings=None, maxNum=-1):
   return ret
 
 def search(args):
+  """
+  Default function for the search command line option.
+  Search a process's memory for a specific Structure.
+  Returns findings in pickled or text format.
+  
+  See the command line --help .
+  """
   log.debug('args: %s'%args)
-  structType = getKlass(args.structType)
+  structType = getKlass(args.structName)
+  if args.baseOffset:
+    args.baseOffset=int(args.baseOffset,16)
   mappings = MemoryMapper(args).getMappings()
   if args.fullscan:
     targetMapping = mappings
@@ -333,10 +452,17 @@ def search(args):
 
 
 def refresh(args):
+  """
+  Default function for the refresh command line option.
+  Try to map a Structure from a specific offset in memory.
+  Returns it in pickled or text format.
+  
+  See the command line --help .
+  """
   log.debug(args)
 
   addr=int(args.addr,16)
-  structType=getKlass(args.structType)
+  structType=getKlass(args.structName)
 
   mappings = MemoryMapper(args).getMappings()
   finder = StructFinder(mappings)
@@ -368,17 +494,6 @@ def refresh(args):
       print pickle.dumps(d)
   return instance,validated
 
-def test():
-  import subprocess
-  cmd_line=['python', 'abouchet.py', 'refresh', '2442', 'ctypes_openssh.session_state', '0xb822a268']
-  p = subprocess.Popen(cmd_line, stdin=None, stdout=subprocess.PIPE, close_fds=True )
-  p.wait()
-  instance=p.stdout.read()
-  instance=eval(instance)
-  return instance
-
-def devnull(arg, **args):
-  return
 
 def main(argv):
   
@@ -386,12 +501,14 @@ def main(argv):
   opts = parser.parse_args(argv)
   if opts.debug:
     logging.basicConfig(level=logging.DEBUG)
+  elif opts.quiet:
+    logging.basicConfig(level=logging.ERROR)    
   else:
     logging.basicConfig(level=logging.INFO)
   try:
     opts.func(opts)
   except ImportError,e:
-    log.error('Struct type does not exists.')
+    log.error('Structure type does not exists.')
     log.error('sys.path is %s'%sys.path)
     print e
 
