@@ -27,30 +27,52 @@ if ctypes.Structure.__name__ == 'Structure':
 class _book(object):
   modules = set()
   classes = dict()
-  refs = list()
+  refs = dict()
   def __init__(self):
     pass
   def addModule(self, mod):
     self.modules.add(mod)
   def addClass(self,cls):
     self.classes[ctypes.POINTER(cls)] = cls
-  def addRef(self,obj):
-    self.refs.append(obj)
+  def addRef(self,obj, typ, addr):
+    self.refs[(typ,addr)]=obj
   def getModules(self):
     return set(self.modules)
   def getClasses(self):
     return dict(self.classes)
+  def getRef(self,typ,addr):
+    return self.refs[(typ,addr)]
   def isRegisteredType(self, typ):
     return typ in self.classes.values()
 
+    
 # central model book register
 __book = _book()
 
-def keepRef(obj):
+def printRefs():
+  l=[(typ,obj,addr) for ((typ,addr),obj) in __book.refs.items()]
+  for i in l:
+    print l
+
+def hasRef(typ,origAddr):
+  return (typ,origAddr) in __book.refs
+
+def getRef(typ,origAddr):
+  if (typ,origAddr) in __book.refs:
+    return __book.getRef(typ,origAddr)
+  return None
+
+def keepRef(obj,typ=None,origAddr=None):
   ''' Sometypes, your have to cast a c_void_p, You can keep ref in Ctypes object, 
     they might be transient (if obj == somepointer.contents).'''
-  #__refs.append(obj)
-  __book.addRef(obj)
+  if (typ,origAddr) in __book.refs:
+    if origAddr is not None:
+      origAddr=hex(origAddr)
+    else:
+      origAddr='None'
+    if typ is not None:
+      log.warning('references already in cache %s/%s'%(typ,origAddr))
+  __book.addRef(obj,typ,origAddr)
   return
 
 def register(klass):
@@ -109,6 +131,16 @@ def getaddress(obj):
     #print 'object pointer is null'
     return 0  
 
+def container_of(memberaddr, typ, membername):
+  '''      const typeof( ((typ *)0)->member ) *__mptr = (ptr);    
+        (type *)( (char *)__mptr - offsetof(type,member) );}) '''
+  return typ.from_address( memberaddr - offsetof(typ, membername) )
+
+def offsetof(typ, membername):
+  T=typ()
+  return ctypes.addressof(  getattr(T,membername) ) - ctypes.addressof(T)
+
+
 ''' MISSING
 d 	double 	float 	8 	(4)
 p 	char[] 	string 	  	 
@@ -388,7 +420,7 @@ class LoadableMembers(ctypes.Structure):
       # all case, 
       _attrType=None
       if attrtype not in self.classRef:
-        log.debug("I can't know the size of the basic type behind the %s pointer, it's a pointer to basic type")
+        log.debug("I can't know the size of the basic type behind the %s pointer, it's a pointer to basic type"%(attrname))
         _attrType=None
       else:
         # test valid address mapping
@@ -428,8 +460,8 @@ class LoadableMembers(ctypes.Structure):
     @returns True if everything has been loaded, False if something went wrong. 
     '''
     if maxDepth == 0:
-      log.warning('Maximum depth reach. Not loading any deeper members.')
-      log.warning('Struct partially LOADED. %s not loaded'%(self.__class__.__name__))
+      log.debug('Maximum depth reach. Not loading any deeper members.')
+      log.debug('Struct partially LOADED. %s not loaded'%(self.__class__.__name__))
       return True
     maxDepth-=1
     log.debug('%s loadMembers'%(self.__class__.__name__))
@@ -451,7 +483,7 @@ class LoadableMembers(ctypes.Structure):
         if not self._loadMember(attr,attrname,attrtype,mappings, maxDepth):
           return False
       except ValueError, e:
-        log.error( 'maxDepath was %d'% maxDepth)
+        log.error( 'maxDepth was %d'% maxDepth)
         raise e
 
     log.debug('%s END loadMembers ----------------'%(self.__class__.__name__))
@@ -526,9 +558,18 @@ class LoadableMembers(ctypes.Structure):
         # big BUG Badaboum, why did pointer changed validity/value ?
         log.warning("%s %s not loadable 0x%lx but VALID "%(attrname, attr,attr_obj_address ))
         return True
+      ref=getRef(_attrType,attr_obj_address)
+      if ref:
+        log.debug("%s %s loading from references cache %s/0x%lx"%(attrname,attr,_attrType,attr_obj_address ))
+        attr.contents = ref
+        return True
       log.debug("%s %s loading from 0x%lx (is_valid_address: %s)"%(attrname,attr,attr_obj_address, memoryMap ))
       ##### VALID INSTR.
+      if attrname == 'real_parent':
+        print 'real_parent',__book.refs
       attr.contents=_attrType.from_buffer_copy(memoryMap.readStruct(attr_obj_address, _attrType ))
+      # save that ref and original addr so we dont need to recopy it later
+      keepRef( attr.contents, _attrType,attr_obj_address)
       #####
       log.debug("%s %s loaded memcopy from 0x%lx to 0x%lx"%(attrname, attr, attr_obj_address, (getaddress(attr))   ))
       # recursive validation checks on new struct
@@ -555,9 +596,6 @@ class LoadableMembers(ctypes.Structure):
     s=''
     if isStructType(attr):
       s=prefix+'"%s": {\t%s%s},\n'%(field, attr.toString(prefix+'\t'),prefix )  
-    #elif isBasicTypeArrayType(attr):
-    #  #s=prefix+'"%s": %s,\n'%(field, array2bytes(attr) )  
-    #  s='['+','.join(["%lx"%(val) for val in attr ])
     elif isBasicTypeArrayType(attr): ## array of something else than int
       s=prefix+'"%s": b%s,\n'%(field, repr(array2bytes(attr)) )  
       #s=prefix+'"%s" :['%(field)+','.join(["0x%lx"%(val) for val in attr ])+'],\n'
@@ -577,7 +615,9 @@ class LoadableMembers(ctypes.Structure):
       else:
         # we can read the pointers contents # if isBasicType(attr.contents): ?  # if isArrayType(attr.contents): ?
         contents=attr.contents
-        if isStructType(contents):
+        if type(self) == type(contents):
+          s=prefix+'"%s": { #(0x%lx) -> %s\n%s},\n'%(field, getaddress(attr), type(attr.contents), prefix) # use struct printer
+        elif isStructType(contents): # do not enter in lists
           s=prefix+'"%s": { #(0x%lx) -> %s%s},\n'%(field, getaddress(attr), attr.contents.toString(prefix+'\t'),prefix) # use struct printer
         elif isPointerType(contents):
           s=prefix+'"%s": { #(0x%lx) -> %s%s},\n'%(field, getaddress(attr), self._attrToString(attr.contents, None, None, prefix+'\t'), prefix ) # use struct printer
@@ -585,8 +625,10 @@ class LoadableMembers(ctypes.Structure):
           s=prefix+'"%s": { #(0x%lx) -> %s\n%s},\n'%(field, getaddress(attr), attr.contents, prefix) # use struct printer
     elif isCStringPointer(attr):
       s=prefix+'"%s": "%s" , #(CString)\n'%(field, attr.string)  
-    else:
-      s=prefix+'"%s": %s, # DEFAULT toString\n'%(field, repr(attr) )  
+    elif isBasicType(attr): # basic, ctypes.* !Structure/pointer % CFunctionPointer?
+      s=prefix+'"%s": %s, \n'%(field, repr(attr) )  
+    else: # wtf ?
+      s=prefix+'"%s": %s, # Unknown/bug DEFAULT repr\n'%(field, repr(attr) )  
     return s
 
   def __str__(self):
@@ -610,10 +652,9 @@ class LoadableMembers(ctypes.Structure):
           s+='%s (@0x%lx) : 0x%lx\n'%(field,ctypes.addressof(attr), getaddress(attr) )   # only print address/null
         elif not is_address_local(attr) :
           s+='%s (@0x%lx) : 0x%lx (FIELD NOT LOADED)\n'%(field,ctypes.addressof(attr), getaddress(attr) )   # only print address in target space
+        elif type(self) == type(attr.contents): # do not recurse in lists
+          s+='%s (@0x%lx) : (0x%lx) -> {%s}\n'%(field, ctypes.addressof(attr), getaddress(attr), repr(attr.contents) ) # use struct printer
         else:
-          # we can read the pointers contents
-          # if isBasicType(attr.contents): ?
-          # if isArrayType(attr.contents): ?
           s+='%s (@0x%lx) : (0x%lx) -> {%s}\n'%(field, ctypes.addressof(attr), getaddress(attr), attr.contents) # use struct printer
       elif isCStringPointer(attr):
         s+='%s (@0x%lx) : %s (CString) \n'%(field,ctypes.addressof(attr), attr.string)  
@@ -631,6 +672,10 @@ class LoadableMembers(ctypes.Structure):
     #log.info("%s %s %s_py"%(self.__class__.__module__, sys.modules[self.__class__.__module__], self.__class__.__name__) )
     my_class=getattr(sys.modules[self.__class__.__module__],"%s_py"%(self.__class__.__name__) )
     my_self=my_class()
+    #keep ref
+    if hasRef(my_class, ctypes.addressof(self) ):
+      return getRef(my_class, ctypes.addressof(self) )
+    keepRef(my_self, my_class, ctypes.addressof(self) )
     _fieldsTuple = [ (f[0],f[1]) for f in self._fields_] 
     for field,typ in _fieldsTuple:
       attr=getattr(self,field)
@@ -658,6 +703,10 @@ class LoadableMembers(ctypes.Structure):
       else:
         contents=attr.contents
         if isStructType(contents) :
+          attr_py_class = getattr(sys.modules[contents.__class__.__module__],"%s_py"%(contents.__class__.__name__) )
+          cache = getRef(attr_py_class, getaddress(attr) )
+          if cache:
+            return cache
           obj=contents.toPyObject()
         elif isPointerType(contents):
           obj=self._attrToPyObject(contents,None,None)
@@ -683,24 +732,26 @@ class pyObj(object):
     - findCtypes(self) : checks if a ctypes is to be found somewhere is the object.
                       Useful to check if the object can be pickled.
   '''
-  def toString(self, prefix=''):
+  def toString(self, prefix='',maxDepth=10):
+    if maxDepth < 0:
+      return '#(- not printed by Excessive recursion - )'
     s='{\n'
     for attrname,typ in self.__dict__.items():
       attr = getattr(self, attrname)
-      s += "%s%s: %s\n"%( prefix, attrname, self._attrToString(attr, attrname, typ, prefix+'\t') )
+      s += "%s%s: %s\n"%( prefix, attrname, self._attrToString(attr, attrname, typ, prefix+'\t', maxDepth=maxDepth-1) )
     s+='}'
     return s
 
-  def _attrToString(self, attr, attrname, typ, prefix ):
+  def _attrToString(self, attr, attrname, typ, prefix, maxDepth):
     s=''
     if type(attr) is tuple or type(attr) is list:
       for i in xrange(0,len(attr)):
-        s += '%s,'%(self._attrToString(attr[i], i ,None, prefix+'\t' ) )
+        s += '%s,'%(self._attrToString(attr[i], i ,None, prefix+'\t', maxDepth) )
       s = "[%s],"%(s)
     elif not hasattr(attr,'__dict__'):
       s = '%s,'%( repr(attr) )
     elif  isinstance( attr , pyObj):
-      s = '%s,'%( attr.toString(prefix) )
+      s = '%s,'%( attr.toString(prefix,maxDepth) )
     else:
       s = '%s,'%(repr(attr) )
       #print 'ELSE type: %s %s'%(type(attr), type(type(attr)) )
@@ -739,7 +790,7 @@ class pyObj(object):
 
   def __iter__(self):
     ''' iterate on a instance's type's _fields_ members following the original type field order '''
-    for k,typ in self._ctype_._fields_:
+    for k,typ in [ (f[0],f[1]) for f in self._ctype_._fields_]:
       v = getattr(self,k)
       yield (k,v,typ)
     pass
